/*

  sshgenvalist.h

  Author: Mika Kojo <mkojo@ssh.fi>

  Copyright (c) 1999 SSH Communications Security, Finland
  All rights reserved.

  Created Tue Jun 13 19:13:31 2000.

  */

#ifndef SSH_GENVALIST_H
#define SSH_GENVALIST_H

/* WARNING: This file was generated using genvalist-data.pl. The 
   contents of this file should not be modified, instead make the 
   required modifications to files:

     genvalist-data.pl
       The Perl program which combines the pieces and generated this 
       file. 

     genvalist-data.txt
       A plain text file which contains extra include file descriptions 
       and supported type listing. 

     genvalist-raw.c
       File that contains the actual C implementation of required 
       functions. Part of this file is left unspecified, as the 
       Perl program fills these gaps. This file is used to generate
       the file sshgenvalist.c.

     genvalist-raw.h
       The raw header file. The Perl program uses this to generate 
       sshgenvalist.h. 
*/

/* "Generic" vararg list manipulation. Explicitly this code can handle
   varargs of form

   TAG, [args], TAG, ..., ENDTAG

   so it is not entirely generic. */

/**********************************************************************/
/* Following definitions are of little interest to the API user, please
   ignore them for most part.

   CAUTION: The application should NOT use the knowledge of the
            internals of the implementation. The implementation
            details may change without a notice, and may even
            be different on different platforms. 

   */

#include "sshincludes.h"
#include "sshcrypt.h"
#include "sshmp.h"
#include "sshbuffer.h"

typedef enum { SSH_GEN_VALIST_EMPTY = 0 } SshGenVaListEnum;

typedef struct SshGenVaListEntryRec *SshGenVaListEntries;

typedef enum
{
  SSH_GEN_VALIST_INVALID = 0,
  SSH_GEN_VALIST_Boolean,
  SSH_GEN_VALIST_Boolean_p,
  SSH_GEN_VALIST_Boolean_pp,
  SSH_GEN_VALIST_unsigned_char,
  SSH_GEN_VALIST_unsigned_char_p,
  SSH_GEN_VALIST_unsigned_char_pp,
  SSH_GEN_VALIST_char,
  SSH_GEN_VALIST_char_p,
  SSH_GEN_VALIST_char_pp,
  SSH_GEN_VALIST_unsigned_int,
  SSH_GEN_VALIST_unsigned_int_p,
  SSH_GEN_VALIST_unsigned_int_pp,
  SSH_GEN_VALIST_int,
  SSH_GEN_VALIST_int_p,
  SSH_GEN_VALIST_int_pp,
  SSH_GEN_VALIST_void_p,
  SSH_GEN_VALIST_void_pp,
  SSH_GEN_VALIST_SshMPInt,
  SSH_GEN_VALIST_SshMPInt_p,
  SSH_GEN_VALIST_SshBuffer,
  SSH_GEN_VALIST_SshBuffer_p,
  SSH_GEN_VALIST_SshGenVaListEnum,
  SSH_GEN_VALIST_SshGenVaListEnum_p,
  SSH_GEN_VALIST_SshGenVaListEnum_pp,
  SSH_GEN_VALIST_END = 65535
} SshGenVaListType;

/* Generated data structure for entries. */
struct SshGenVaListEntryRec
{
  SshGenVaListType type;
  union {
    Boolean data_Boolean;
    Boolean * data_Boolean_p;
    Boolean ** data_Boolean_pp;
    unsigned char data_unsigned_char;
    unsigned char * data_unsigned_char_p;
    unsigned char ** data_unsigned_char_pp;
    char data_char;
    char * data_char_p;
    char ** data_char_pp;
    unsigned int data_unsigned_int;
    unsigned int * data_unsigned_int_p;
    unsigned int ** data_unsigned_int_pp;
    int data_int;
    int * data_int_p;
    int ** data_int_pp;
    void * data_void_p;
    void ** data_void_pp;
    SshMPInt data_SshMPInt;
    SshMPInt * data_SshMPInt_p;
    SshBuffer data_SshBuffer;
    SshBuffer * data_SshBuffer_p;
    SshGenVaListEnum data_SshGenVaListEnum;
    SshGenVaListEnum * data_SshGenVaListEnum_p;
    SshGenVaListEnum ** data_SshGenVaListEnum_pp;
  } c;
};

struct SshGenVaListRec
{
  unsigned int max, used, pos;
  SshGenVaListEntries entry;
};

/**********************************************************************/
/* Following is the main API definition. */

/* The definition data structures. */

typedef struct SshGenVaListDefRec *SshGenVaListDef;
typedef struct SshGenVaListDefRec  SshGenVaListDefStruct;
typedef struct SshGenVaListRec    *SshGenVaList;

struct SshGenVaListDefRec
{
  /* We cannot use here e.g. SshGenVaListEnum as this could cause
     warnings and problem, probably. */
  unsigned int tag;
  const unsigned char *types;
};

/* The basic principle of this library is to take an opened "va_list",
   and convert it into a data structure that is more robust and
   platform independend.

   That is, the caller must first call "va_start", before calling this
   function. This is similar to vsprintf, for example.

   Caller also needs a definition of the "syntax" of the vararg. This
   can be accomplished by defining a constant data structure to which
   "def" points to. The "def_num_elem" gives the number of elements
   in the definition table. 

   The returned "SshGenVaList" is allocated and can be stored for later
   use, if wished. 
   */
SshGenVaList ssh_gen_valist_alloc(SshGenVaListDef def,
                                  unsigned int def_num_elem,
                                  va_list ap);

/* This routine frees the allocated "SshGenVaList" data structure. This
   must be called once, and only once, for each "SshGenVaList". */
void ssh_gen_valist_free(SshGenVaList vl);


/* A simple way to access the contents of the "SshGenVaList". The
   syntax of this macro is as follows:

   Assume you know that type "Boolean" resides as the next in the
   vararg. Then writing

   myboolean = SSH_GEN_VALIST_ARG(vl, Boolean);

   obtains the next argument. Observe that the macro uses quite strict
   checking, which is of course only good for debugging purposes.

   The anomaly, which may cause some discomfort for some, is that you
   can NOT use types such as "unsigned char" or "Boolean **" directly.
   Instead you have to, by hand, use following simple canonicalization
   and conversion algorithm:

   Input:
     A C type, say, X.
   
   Whitespaces to underscore:
     Replace all continuous whitespace blocks by a "_" (underscore).
     That is, long continuous sequences of whitespace give only one (1)
     underscore character. Remove possible underscores from beginning and
     end. Remove underscores separating special characters (such as
     stars (*)). 

     Example: If X = "  unsigned char  ", you must by hand convert it to
              X = "unsigned_char".

   Stars to p's:
     Now in X may still appear characters that are not valid. Indeed,
     the only characters that are interesting to us are stars (*). We
     require you to convert all stars (*) into p's (p). In the resulting
     string there should be equally many p's as there were stars in the
     original. 

     Example: If X = "  unsigned char ** ", you must by hand convert it to
              X = "unsigned_char_pp". 
     
   After this manipulation you can write SSH_GEN_VALIST_ARG(vl, X).

   Example: If X = "   unsigned char **  ", we have
            SSH_GEN_VALIST_ARG(vl, unsigned_char_pp);

   This encoding method does work for all supported types, and if the
   type is not supported by the library you will get a compilation time
   error.


   Some special cases:

   * when reading an enumtype use "SshGenVaListEnum".
   * this library implements only finitely many types (and thus you
     cannot expect "char ******...*" to work, even if your compiler would
     allow it). 
   
*/

#define SSH_GEN_VALIST_ARG(valist, type) \
  (SSH_ASSERT((valist)->pos < (valist)->max), \
   SSH_ASSERT((valist)->entry[(valist)->pos].type == SSH_GEN_VALIST_##type), \
   (valist)->entry[(valist)->pos++].c.data_##type)

/* Management of the vararg list. */

/* Definitions to move to start and end of the vararg list. */
#define SSH_GEN_VALIST_REWIND_START -0x7fff
#define SSH_GEN_VALIST_REWIND_END    0x8000

/* This function can be used to move in the "SshGenVaList". Negative
   numbers move "backwards" and positive "forwards". You cannot move
   past first nor last of the valid argument positions (in case of
   reference outside the possible range of indexes closest will be
   selected). */
void ssh_gen_valist_rewind(SshGenVaList vl, int steps);

#endif /* SSH_GENVALIST_H */
