/*

  t-float.c

  Author: Mika Kojo <mkojo@ssh.fi>

  Copyright (c) 2000 SSH Communications Security Corp
  All rights reserved.

  Created Mon Jan  8 03:40:34 2001.

  */

/* Simple testing for the float library. */

#include "sshincludes.h"
#include "sshmp.h"

void print(char *str, SshMPFloat op)
{
#if 1
  char *t;
  t = ssh_mprf_get_str(op);
  fprintf(stderr, "%s = %s\n", str, t);
  ssh_xfree(t);
#else
  printf("%s = ", str);
  ssh_mprf_dump(op);
  printf("\n");
#endif
}

void printr(char *str, SshMPRational op)
{
  char *a,*b;
  a = ssh_mprz_get_str(&op->n, 10);
  b = ssh_mprz_get_str(&op->d, 10);
  fprintf(stderr, "%s = %s/%s\n", str, a, b);
  ssh_xfree(a); ssh_xfree(b);
}

void test1(void)
{
  SshMPFloatStruct a, b, c, d, r;
  SshMPRationalStruct rt;
  
  ssh_mprf_init_with_prec(&a, 20);
  ssh_mprf_init_with_prec(&b, 20);
  ssh_mprf_init_with_prec(&c, 20);
  ssh_mprf_init_with_prec(&d, 20);
  ssh_mprf_init_with_prec(&r, 20);

  ssh_mprq_init(&rt);
  
  ssh_mprf_rand(&r);
  ssh_mprf_set_ui(&a, ssh_rand() % 10000);
  ssh_mprf_set_ui(&b, (ssh_rand() % 10000) + 1);

  print("a", &a);
  print("b", &b);
  print("r", &r);
  
  ssh_mprf_add(&c, &a, &b);
  print("a + b", &c);

  ssh_mprf_sub(&c, &a, &b);
  print("a - b", &c);

  ssh_mprf_mul_2exp(&c, &r, 20);
  print("r * 2^20", &c);

  ssh_mprf_mod1(&d, &c);
  print("r * 2^20 mod 1", &d);

  ssh_mprf_floor(&d, &c);
  print("floor r * 2^20", &d);
  
  ssh_mprf_div_2exp(&c, &b, 10);
  print("b / 2^10", &c);
  ssh_mprf_add(&c, &a, &c);
  print("a + (b / 2^10)", &c);
  
  ssh_mprf_mul(&c, &a, &b);
  print("a * b", &c);
  
  ssh_mprf_div(&c, &a, &b);
  print("a / b", &c);

  ssh_mprf_sqrt(&c, &a);
  print("sqrt a", &c);

  ssh_mprf_square(&c, &c);
  print("(sqrt a)^2", &c);

  ssh_mprf_pi(&c);
  print("pi", &c);

  ssh_mprf_ln2(&c);
  print("ln 2", &c);

  ssh_mprf_pi(&c);
  ssh_mprf_ln2(&d);
  ssh_mprf_add(&c, &c, &d);
  ssh_mprf_sqrt(&c, &c);
  print("sqrt(pi + ln2)", &c);

  ssh_mprf_set_ui(&c, 1);
  ssh_mprf_exp(&c, &c);
  print("exp(1)", &c);

  ssh_mprf_set_ui(&c, 2);
  ssh_mprf_ln(&c, &c);
  print("ln(2)", &c);

  ssh_mprf_set_ui(&c, 2);
  ssh_mprf_log(&c, &c);
  print("log(2)", &c);
  
  ssh_mprf_log(&c, &r);
  print("log(r)", &c);
  
  ssh_mprf_set_ui(&c, 2);
  ssh_mprf_pow(&c, &c, &r);
  print("2^r", &c);

  ssh_mprf_set_ui(&c, 1);
  ssh_mprf_atan(&c, &c);
  print("atan(1)", &c);

  ssh_mprf_pi(&c);
  ssh_mprf_div_2exp(&c, &c, 2);
  ssh_mprf_cos(&d, &c);
  print("cos pi/4", &d);

  ssh_mprf_neg(&c, &c);
  ssh_mprf_sin(&d, &c);
  print("sin -pi/4", &d);

  ssh_mprf_pi(&c);
  ssh_mprf_set_ui(&d, 12);
  ssh_mprf_div(&c, &c, &d);
  ssh_mprf_neg(&c, &c);
  ssh_mprf_tan(&d, &c);
  print("tan -pi/12", &d);

  ssh_mprf_pi(&c);
  ssh_mprf_get_mprq(&rt, &c);
  printr("approx pi", &rt);
  printf("Doubled %g\n", ssh_mprq_get_d(&rt));
  printf("Double %g/%g\n", ssh_mprz_get_d(&rt.n),
         ssh_mprz_get_d(&rt.d));

  ssh_mprf_set_mprq(&d, &rt);
  print("approx pi", &d);
  ssh_mprf_sub(&c, &c, &d);
  print("diff     ", &c);

  ssh_mprf_clear(&a);
  ssh_mprf_clear(&b);
  ssh_mprf_clear(&c);
  ssh_mprf_clear(&d);

  ssh_mprq_clear(&rt);
}

void test2(void)
{
  SshMPFloatStruct a, b, c, d;
  int i;
  
  ssh_mprf_init(&a);
  ssh_mprf_init(&b);
  ssh_mprf_init(&c);
  ssh_mprf_init(&d);
  
  for (i = 0; i < 20; i++)
    {
      int rv;
      ssh_mprf_set_ui(&a, ssh_rand() % 100000);
      ssh_mprf_set_ui(&b, ssh_rand() % 100000);

      ssh_mprf_div_2exp(&a, &a,
                        (ssh_rand() % 100) * ((ssh_rand() & 1) ? 1 : -1));
      ssh_mprf_div_2exp(&b, &b,
                        (ssh_rand() % 100) * ((ssh_rand() & 1) ? 1 : -1));

      rv = ssh_mprf_cmp(&a, &b);
      print("a", &a);
      print("b", &b);
      printf("==> %s\n", (rv == 0) ? "equal" :
             (rv < 0) ? "less than" : "greater than");
    }
  
  ssh_mprf_clear(&a);
  ssh_mprf_clear(&b);
  ssh_mprf_clear(&c);
  ssh_mprf_clear(&d);
}

void mand(unsigned int xsize, unsigned int ysize,
          SshMPFloat xmin, SshMPFloat ymin,
          SshMPFloat xmax, SshMPFloat ymax,
          unsigned int itr,
          unsigned int *ptr)
{
  SshMPFloatStruct si, sr, ar, ai, fr, fi, br, bi, ax, ay, t;
  unsigned int x, y, i, verb;

  ssh_mprf_init_inherit_prec(&si, xmin);
  ssh_mprf_init_inherit_prec(&sr, xmin);
  ssh_mprf_init_inherit_prec(&ar, xmin);
  ssh_mprf_init_inherit_prec(&ai, xmin);
  ssh_mprf_init_inherit_prec(&fr, xmin);
  ssh_mprf_init_inherit_prec(&fi, xmin);
  ssh_mprf_init_inherit_prec(&br, xmin);
  ssh_mprf_init_inherit_prec(&bi, xmin);
  ssh_mprf_init_inherit_prec(&ax, xmin);
  ssh_mprf_init_inherit_prec(&ay, xmin);
  ssh_mprf_init_inherit_prec(&t,  xmin);

  /* Compute the step size. */
  ssh_mprf_sub(&ax, xmax, xmin);
  ssh_mprf_sub(&ay, ymax, ymin);
  ssh_mprf_set_ui(&t, xsize);
  print("t", &t);
  ssh_mprf_div(&ax, &ax, &t);
  ssh_mprf_set_ui(&t, ysize);
  print("t", &t);
  ssh_mprf_div(&ay, &ay, &t);

  print("xmin", xmin);
  print("ymin", ymin);
  print("xmax", ymax);
  print("ymax", ymax);
  print("ax", &ax);
  print("ay", &ay);
  
  /* Set up the iteration. */
  ssh_mprf_set(&si, ymin);

  for (y = 0; y < ysize; y++)
    {
      ssh_mprf_set(&sr, xmin);
      for (x = 0; x < xsize; x++)
        {
          ssh_mprf_square(&ar, &sr);
          ssh_mprf_square(&ai, &si);
          ssh_mprf_set(&fr, &sr);
          ssh_mprf_set(&fi, &si);

          verb = ((x == 39) && (y == 20)) ? 1 : 0; 

          if (verb) print("sr", &sr);
          if (verb) print("si", &si);
          
          for (i = 0; i < itr; i++)
            {
              ssh_mprf_add(&t, &ar, &ai);
              if (verb) print("ar", &ar);
              if (verb) print("ai", &ai);
              if (verb) print("t",  &t);
              if (ssh_mprf_cmp_ui(&t, 4) >= 0)
                break;

              if (verb) print("br", &br);
              if (verb) print("bi", &bi);

              /* br = ar - ai + sr */
              ssh_mprf_sub(&br, &ar, &ai);
              ssh_mprf_add(&br, &br, &sr);

              /* bi = fr*fi*2 + si */
              ssh_mprf_mul(&bi, &fr, &fi);
              ssh_mprf_mul_2exp(&bi, &bi, 1);
              ssh_mprf_add(&bi, &bi, &si);

              if (verb) print("br", &br);
              if (verb) print("bi", &bi);

              ssh_mprf_square(&ar, &br);
              ssh_mprf_square(&ai, &bi);
              ssh_mprf_set(&fr, &br);
              ssh_mprf_set(&fi, &bi);
            }

          ptr[x + y*xsize] = itr - i;
          ssh_mprf_add(&sr, &sr, &ax);
        }
      ssh_mprf_add(&si, &si, &ay);
    }
  
  ssh_mprf_clear(&si);
  ssh_mprf_clear(&sr);
  ssh_mprf_clear(&ar);
  ssh_mprf_clear(&ai);
  ssh_mprf_clear(&fr);
  ssh_mprf_clear(&fi);
  ssh_mprf_clear(&br);
  ssh_mprf_clear(&bi);
  ssh_mprf_clear(&ax);
  ssh_mprf_clear(&ay);
  ssh_mprf_clear(&t);

  return;
}


void mandd(unsigned int xsize, unsigned int ysize,
           double xmin, double ymin,
           double xmax, double ymax,
           unsigned int itr,
           unsigned int *ptr)
{
  double si, sr, ar, ai, fr, fi, br, bi, ax, ay, t;
  unsigned int x, y, i, verb;

  /* Compute the step size. */
  ax = xmax - xmin;
  ay = ymax - ymin;
  ax = ax / (double)xsize;
  ay = ay / (double)ysize;

  printf("xmin = %g\n", xmin);
  printf("ymin = %g\n", ymin);
  printf("xmax = %g\n", ymax);
  printf("ymax = %g\n", ymax);
  printf("ax = %g\n", ax);
  printf("ay = %g\n", ay);
  
  /* Set up the iteration. */
  si = ymin;

  for (y = 0; y < ysize; y++)
    {
      sr = xmin;
      for (x = 0; x < xsize; x++)
        {
          ar = sr*sr;
          ai = si*si;
          fr = sr;
          fi = si;

          verb = ((x == 39) && (y == 20)) ? 1 : 0; 

          if (verb) printf("sr %g si %g\n", sr, si);
          
          for (i = 0; i < itr; i++)
            {
              t = ar + ai;
              if (verb) printf("ar %g ai %g t %g\n", ar, ai, t);
              if (t >= 4.0)
                break;

              if (verb) printf("%g %g\n", br, bi);
              
              /* br = ar - ai + sr */
              br = ar - ai + sr;

              /* bi = fr*fi*2 + si */
              bi = fr*fi*2 + si;

              if (verb) printf("%g %g\n", br, bi);

              ar = br*br;
              ai = bi*bi;
              fr = br;
              fi = bi;
            }

          ptr[x + y*xsize] = itr - i;
          sr = sr + ax;
        }
      si = si + ay;
    }

  return;
}


void render(unsigned int xsize, unsigned int ysize,
            unsigned int *ptr)
{
  unsigned char *color_str =
    "  .,`-^;:*?#";
  size_t color_len = strlen(color_str);
  int x, y;

  for (y = 0; y < ysize; y++)
    {
      for (x = 0; x < xsize; x++)
        {
          int t1, t2, p;
          p = x + y*xsize;
          t1 = ptr[p];
          t2 = ptr[p] + 1;

          if (t1 >= color_len) t1 = color_len-1;
          if (t2 >= color_len) t2 = color_len-1;
          
          printf("%c%c", color_str[t1], color_str[t2]);
        }
      printf("\n");
    }
}

void test_mand(void)
{
  SshMPFloatStruct xmin, xmax, ymin, ymax;
  unsigned int *ptr;
#define SX  40
#define SY  24
#define ITR 32
  ptr = ssh_xmalloc(SX * SY * sizeof(unsigned int));

  ssh_mprf_init_with_prec(&xmin, 10);
  ssh_mprf_init_with_prec(&xmax, 10);
  ssh_mprf_init_with_prec(&ymin, 10);
  ssh_mprf_init_with_prec(&ymax, 10);

  ssh_mprf_set_si(&xmin, -2);
  ssh_mprf_set_si(&xmax,  1);
  ssh_mprf_set_si(&ymin, -2);
  ssh_mprf_set_si(&ymax,  1);

  mand(SX, SY, &xmin, &ymin, &xmax, &ymax, ITR, ptr);

  render(SX, SY, ptr);
  
  ssh_mprf_clear(&xmin);
  ssh_mprf_clear(&xmax);
  ssh_mprf_clear(&ymin);
  ssh_mprf_clear(&ymax);
}

void test_mandd(void)
{
  double xmin, xmax, ymin, ymax;
  unsigned int *ptr;

  ptr = ssh_xmalloc(SX * SY * sizeof(unsigned int));

  xmin = -2;
  xmax = 1;
  ymin = -2;
  ymax = 1;

  mandd(SX, SY, xmin, ymin, xmax, ymax, ITR, ptr);

  render(SX, SY, ptr);
}
































void todouble(int bits, int step)
{
  printf("SshDouble not supported.\n");
}


int main(int ac, char *av[])
{
  ssh_rand_seed(ssh_time());
  test1(); /* test2(); */
  todouble(1000, 100);
  printf("\nMPFLOAT: \n");
  test_mand();
  printf("\nDOUBLE: \n");
  test_mandd();
  return 0;
}
